@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 #
-#  Copyright (C) 2017 Codethink Limited
+#  Copyright (C) 2017-2018 Codethink Limited
 #
 #  This program is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU Lesser General Public
@@ -20,9 +20,9 @@
 
 import multiprocessing
 import os
+import signal
 import string
 import tempfile
-import signal
 
 from .. import _ostree, _signals, utils
 from .._exceptions import ArtifactError
@@ -45,16 +45,19 @@ def buildref(element, key):
         for x in element.normal_name
     ])
 
+    if key is None:
+        raise ArtifactError('Cache key missing')
+
     # assume project and element names are not allowed to contain slashes
     return '{0}/{1}/{2}'.format(project.name, element_name, key)
 
 
-# An ArtifactCache manages artifacts in an OSTree repository
+# An OSTreeCache manages artifacts in an OSTree repository
 #
 # Args:
 #     context (Context): The BuildStream context
 #     project (Project): The BuildStream project
-#     enable_push (bool): Whether pushing is allowed
+#     enable_push (bool): Whether pushing is allowed by the platform
 #
 # Pushing is explicitly disabled by the platform in some cases,
 # like when we are falling back to functioning without using
@@ -62,44 +65,33 @@ def buildref(element, key):
 #
 class OSTreeCache(ArtifactCache):
 
-    def __init__(self, context, project, enable_push):
-        super().__init__(context, project)
+    def __init__(self, context, enable_push):
+        super().__init__(context)
 
         self.enable_push = enable_push
 
         ostreedir = os.path.join(context.artifactdir, 'ostree')
         self.repo = _ostree.ensure(ostreedir, False)
 
-        self.push_url = None
-        self.pull_url = None
+        self.push_urls = {}
+        self.pull_urls = {}
+        self._remote_refs = {}
+        self._has_fetch_remotes = False
+        self._has_push_remotes = False
 
-        self._remote_refs = None
+    def has_fetch_remotes(self):
+        return self._has_fetch_remotes
 
-    def initialize_remote(self):
-        if self.url is not None:
-            if self.url.startswith('ssh://'):
-                self.push_url = self.url
-                try:
-                    # Contact the remote cache.
-                    self.pull_url = initialize_push_connection(self.push_url)
-                except PushException as e:
-                    raise ArtifactError("BuildStream did not connect succesfully "
-                                        "to the shared cache: {}".format(e))
-            elif self.url.startswith('http://') or self.url.startswith('https://'):
-                self.push_url = None
-                self.pull_url = self.url
-            elif self._local:
-                self.push_url = self.url
-                self.pull_url = self.url
-            else:
-                raise ArtifactError("Unsupported URL scheme: {}".format(self.url))
-
-            _ostree.configure_remote(self.repo, self.remote, self.pull_url)
-
-    def can_push(self):
-        if self.enable_push:
-            return (not self._offline or self._local) and self.push_url is not None
-        return False
+    def has_push_remotes(self, *, element=None):
+        if not self._has_push_remotes:
+            # No project has push remotes
+            return False
+        elif element is None:
+            # At least one (sub)project has push remotes
+            return True
+        else:
+            # Check whether the specified element's project has push remotes
+            return len(self.push_urls[element._get_project()]) > 0
 
     # contains():
     #
@@ -116,7 +108,10 @@ class OSTreeCache(ArtifactCache):
         if strength is None:
             strength = _KeyStrength.STRONG if element._get_strict() else _KeyStrength.WEAK
 
-        key = element._get_cache_key(strength)
+        if strength == _KeyStrength.STRONG:
+            key = element._get_strict_cache_key()
+        else:
+            key = element._get_cache_key(strength)
         if not key:
             return False
 
@@ -135,11 +130,15 @@ class OSTreeCache(ArtifactCache):
     # Returns: True if the artifact is in the cache, False otherwise
     #
     def remote_contains_key(self, element, key):
-        if not self._remote_refs:
+        if not self._has_fetch_remotes:
+            return False
+
+        remote_refs = self._remote_refs[element._get_project()]
+        if len(remote_refs) == 0:
             return False
 
         ref = buildref(element, key)
-        return ref in self._remote_refs
+        return ref in remote_refs
 
     # remote_contains():
     #
@@ -156,7 +155,10 @@ class OSTreeCache(ArtifactCache):
         if strength is None:
             strength = _KeyStrength.STRONG if element._get_strict() else _KeyStrength.WEAK
 
-        key = element._get_cache_key(strength)
+        if strength == _KeyStrength.STRONG:
+            key = element._get_strict_cache_key()
+        else:
+            key = element._get_cache_key(strength)
         if not key:
             return False
 
@@ -179,7 +181,7 @@ class OSTreeCache(ArtifactCache):
     # Returns: path to extracted artifact
     #
     def extract(self, element):
-        ref = buildref(element, element._get_cache_key())
+        ref = buildref(element, element._get_strict_cache_key())
 
         # resolve ref to checksum
         rev = _ostree.checksum(self.repo, ref)
@@ -230,7 +232,7 @@ class OSTreeCache(ArtifactCache):
     #
     def commit(self, element, content):
         # tag with strong cache key based on dependency versions used for the build
-        ref = buildref(element, element._get_cache_key_for_build())
+        ref = buildref(element, element._get_cache_key())
 
         # also store under weak cache key
         weak_ref = buildref(element, element._get_cache_key(strength=_KeyStrength.WEAK))
@@ -242,29 +244,24 @@ class OSTreeCache(ArtifactCache):
 
     # pull():
     #
-    # Pull artifact from remote repository.
+    # Pull artifact from one of the configured remote repositories.
     #
     # Args:
     #     element (Element): The Element whose artifact is to be fetched
     #     progress (callable): The progress callback, if any
     #
     def pull(self, element, progress=None):
+        project = element._get_project()
 
-        if self._offline and not self._local:
-            raise ArtifactError("Attempt to pull artifact while offline")
-
-        if self.pull_url.startswith("/"):
-            remote = "file://" + self.pull_url
-        else:
-            remote = self.remote
+        remote_refs = self._remote_refs[project]
 
+        ref = buildref(element, element._get_strict_cache_key())
         weak_ref = buildref(element, element._get_cache_key(strength=_KeyStrength.WEAK))
 
         try:
-            if self.remote_contains(element, strength=_KeyStrength.STRONG):
+            if ref in remote_refs:
                 # fetch the artifact using the strong cache key
-                ref = buildref(element, element._get_cache_key())
-                _ostree.fetch(self.repo, remote=remote,
+                _ostree.fetch(self.repo, remote=remote_refs[ref],
                               ref=ref, progress=progress)
 
                 # resolve ref to checksum
@@ -272,17 +269,17 @@ class OSTreeCache(ArtifactCache):
 
                 # update weak ref by pointing it to this newly fetched artifact
                 _ostree.set_ref(self.repo, weak_ref, rev)
-            elif self.remote_contains(element):
+            elif weak_ref in remote_refs:
                 # fetch the artifact using the weak cache key
-                _ostree.fetch(self.repo, remote=remote,
+                _ostree.fetch(self.repo, remote=remote_refs[weak_ref],
                               ref=weak_ref, progress=progress)
 
                 # resolve weak_ref to checksum
                 rev = _ostree.checksum(self.repo, weak_ref)
 
                 # extract strong cache key from this newly fetched artifact
-                element._cached(recalculate=True)
-                ref = buildref(element, element._get_cache_key_from_artifact())
+                element._update_state()
+                ref = buildref(element, element._get_cache_key())
 
                 # create tag for strong cache key
                 _ostree.set_ref(self.repo, ref, rev)
@@ -293,44 +290,6 @@ class OSTreeCache(ArtifactCache):
             raise ArtifactError("Failed to pull artifact for element {}: {}"
                                 .format(element.name, e)) from e
 
-    # fetch_remote_refs():
-    #
-    # Fetch list of artifacts from remote repository.
-    #
-    def fetch_remote_refs(self):
-        if self.pull_url.startswith("/"):
-            remote = "file://" + self.pull_url
-        elif self.remote is not None:
-            remote = self.remote
-        else:
-            raise ArtifactError("Attempt to fetch remote refs without any pull URL")
-
-        def child_action(repo, remote, q):
-            try:
-                q.put((True, _ostree.list_remote_refs(self.repo, remote=remote)))
-            except OSTreeError as e:
-                q.put((False, str(e)))
-
-        q = multiprocessing.Queue()
-        p = multiprocessing.Process(target=child_action, args=(self.repo, remote, q))
-
-        try:
-
-            # Keep SIGINT blocked in the child process
-            with _signals.blocked([signal.SIGINT], ignore=False):
-                p.start()
-
-            ret, res = q.get()
-            p.join()
-        except KeyboardInterrupt:
-            utils._kill_process_tree(p.pid)
-            raise
-
-        if ret:
-            self._remote_refs = res
-        else:
-            raise ArtifactError("Failed to fetch remote refs: {}".format(res))
-
     # push():
     #
     # Push committed artifact to remote repository.
@@ -345,51 +304,199 @@ class OSTreeCache(ArtifactCache):
     # Raises:
     #   (ArtifactError): if there was an error
     def push(self, element):
+        any_pushed = False
 
-        if self._offline and not self._local:
-            raise ArtifactError("Attempt to push artifact while offline")
+        project = element._get_project()
 
-        if self.push_url is None:
-            raise ArtifactError("The protocol in use does not support pushing.")
+        push_urls = self.push_urls[project]
 
-        ref = buildref(element, element._get_cache_key_from_artifact())
+        if len(push_urls) == 0:
+            raise ArtifactError("Push is not enabled for any of the configured remote artifact caches.")
+
+        ref = buildref(element, element._get_cache_key())
         weak_ref = buildref(element, element._get_cache_key(strength=_KeyStrength.WEAK))
-        if self.push_url.startswith("/"):
-            # local repository
-            push_repo = _ostree.ensure(self.push_url, True)
-            _ostree.fetch(push_repo, remote=self.repo.get_path().get_uri(), ref=ref)
-            _ostree.fetch(push_repo, remote=self.repo.get_path().get_uri(), ref=weak_ref)
 
-            # Local remotes are not really a thing, just return True here
-            return True
+        for push_url in push_urls:
+            any_pushed |= self._push_to_remote(push_url, element, ref, weak_ref)
+
+        return any_pushed
+
+    # _initialize_remote():
+    #
+    # Do protocol-specific initialization necessary to use a given OSTree
+    # remote.
+    #
+    # The SSH protocol that we use only supports pushing so initializing these
+    # involves contacting the remote to find out the corresponding pull URL.
+    #
+    # Args:
+    #     url (str): URL of the remote
+    #
+    # Returns:
+    #     (str, str): the pull URL and push URL for the remote
+    #
+    # Raises:
+    #     ArtifactError: if there was an error
+    def _initialize_remote(self, url):
+        if url.startswith('ssh://'):
+            try:
+                push_url = url
+                pull_url = initialize_push_connection(url)
+            except PushException as e:
+                raise ArtifactError(e) from e
+        elif url.startswith('/'):
+            push_url = pull_url = 'file://' + url
+        elif url.startswith('file://'):
+            push_url = pull_url = url
+        elif url.startswith('http://') or url.startswith('https://'):
+            push_url = None
+            pull_url = url
         else:
-            # Push over ssh
-            #
-            with utils._tempdir(dir=self.context.artifactdir, prefix='push-repo-') as temp_repo_dir:
+            raise ArtifactError("Unsupported URL: {}".format(url))
 
-                with element.timed_activity("Preparing compressed archive"):
-                    # First create a temporary archive-z2 repository, we can
-                    # only use ostree-push with archive-z2 local repo.
-                    temp_repo = _ostree.ensure(temp_repo_dir, True)
+        return push_url, pull_url
 
-                    # Now push the ref we want to push into our temporary archive-z2 repo
-                    _ostree.fetch(temp_repo, remote=self.repo.get_path().get_uri(), ref=ref)
-                    _ostree.fetch(temp_repo, remote=self.repo.get_path().get_uri(), ref=weak_ref)
-
-                with element.timed_activity("Sending artifact"), \
-                    element._output_file() as output_file:
-                    try:
-                        pushed = push_artifact(temp_repo.get_path().get_path(),
-                                               self.push_url,
-                                               [ref, weak_ref], output_file)
-                    except PushException as e:
-                        raise ArtifactError("Failed to push artifact {}: {}".format(ref, e)) from e
-
-                return pushed
-
-    # set_offline():
+    # _ensure_remote():
     #
-    # Do not attempt to pull or push artifacts.
+    # Ensure that our OSTree repo has a remote configured for the given URL.
+    # Note that SSH access to remotes is not handled by libostree itself.
     #
-    def set_offline(self):
-        self._offline = True
+    # Args:
+    #     repo (OSTree.Repo): an OSTree repository
+    #     pull_url (str): the URL where libostree can pull from the remote
+    #
+    # Returns:
+    #     (str): the name of the remote, which can be passed to various other
+    #            operations implemented by the _ostree module.
+    #
+    # Raises:
+    #     OSTreeError: if there was a problem reported by libostree
+    def _ensure_remote(self, repo, pull_url):
+        remote_name = utils.url_directory_name(pull_url)
+        _ostree.configure_remote(repo, remote_name, pull_url)
+        return remote_name
+
+    def initialize_remotes(self, *, on_failure=None):
+        remote_specs = self.global_remote_specs
+
+        for project in self.project_remote_specs:
+            remote_specs += self.project_remote_specs[project]
+
+        remote_specs = list(utils._deduplicate(remote_specs))
+
+        remote_results = {}
+
+        # Callback to initialize one remote in a 'multiprocessing' subprocess.
+        #
+        # We cannot do this in the main process because of the way the tasks
+        # run by the main scheduler calls into libostree using
+        # fork()-without-exec() subprocesses. OSTree fetch operations in
+        # subprocesses hang if fetch operations were previously done in the
+        # main process.
+        #
+        def child_action(url, q):
+            try:
+                push_url, pull_url = self._initialize_remote(url)
+                remote = self._ensure_remote(self.repo, pull_url)
+                remote_refs = _ostree.list_remote_refs(self.repo, remote=remote)
+                q.put((None, push_url, pull_url, remote_refs))
+            except Exception as e:
+                q.put((str(e), None, None, None))
+
+        # Kick off all the initialization jobs one by one.
+        #
+        # Note that we cannot use multiprocessing.Pool here because it's not
+        # possible to pickle local functions such as child_action().
+        #
+        q = multiprocessing.Queue()
+        for remote in remote_specs:
+            p = multiprocessing.Process(target=child_action, args=(remote.url, q))
+
+            try:
+
+                # Keep SIGINT blocked in the child process
+                with _signals.blocked([signal.SIGINT], ignore=False):
+                    p.start()
+
+                error, push_url, pull_url, remote_refs = q.get()
+                p.join()
+            except KeyboardInterrupt:
+                utils._kill_process_tree(p.pid)
+                raise
+
+            if error and on_failure:
+                on_failure(remote.url, error)
+            elif error:
+                raise ArtifactError(error)
+            else:
+                if remote.push and push_url:
+                    self._has_push_remotes = True
+                if pull_url:
+                    self._has_fetch_remotes = True
+
+                remote_results[remote.url] = (push_url, pull_url, remote_refs)
+
+        # Prepare push_urls, pull_urls, and remote_refs for each project
+        for project in self.context._get_projects():
+            remote_specs = self.global_remote_specs
+            if project in self.project_remote_specs:
+                remote_specs = list(utils._deduplicate(remote_specs + self.project_remote_specs[project]))
+
+            push_urls = []
+            pull_urls = []
+            _remote_refs = {}
+
+            for remote in remote_specs:
+                # Errors are already handled in the loop above,
+                # skip unreachable remotes here.
+                if remote.url not in remote_results:
+                    continue
+
+                push_url, pull_url, remote_refs = remote_results[remote.url]
+
+                if remote.push:
+                    if push_url:
+                        push_urls.append(push_url)
+                    else:
+                        raise ArtifactError("Push enabled but not supported by repo at: {}".format(remote.url))
+
+                # The specs are deduplicated when reading the config, but since
+                # each push URL can supply an arbitrary pull URL we must dedup
+                # those again here.
+                if pull_url and pull_url not in pull_urls:
+                    pull_urls.append(pull_url)
+
+                # Update our overall map of remote refs with any refs that are
+                # present in the new remote and were not already found in
+                # higher priority ones.
+                remote = self._ensure_remote(self.repo, pull_url)
+                for ref in remote_refs:
+                    if ref not in _remote_refs:
+                        _remote_refs[ref] = remote
+
+            self.push_urls[project] = push_urls
+            self.pull_urls[project] = pull_urls
+            self._remote_refs[project] = _remote_refs
+
+    def _push_to_remote(self, push_url, element, ref, weak_ref):
+        with utils._tempdir(dir=self.context.artifactdir, prefix='push-repo-') as temp_repo_dir:
+
+            with element.timed_activity("Preparing compressed archive"):
+                # First create a temporary archive-z2 repository, we can
+                # only use ostree-push with archive-z2 local repo.
+                temp_repo = _ostree.ensure(temp_repo_dir, True)
+
+                # Now push the ref we want to push into our temporary archive-z2 repo
+                _ostree.fetch(temp_repo, remote=self.repo.get_path().get_uri(), ref=ref)
+                _ostree.fetch(temp_repo, remote=self.repo.get_path().get_uri(), ref=weak_ref)
+
+            with element.timed_activity("Sending artifact"), \
+                element._output_file() as output_file:
+                try:
+                    pushed = push_artifact(temp_repo.get_path().get_path(),
+                                           push_url,
+                                           [ref, weak_ref], output_file)
+                except PushException as e:
+                    raise ArtifactError("Failed to push artifact {}: {}".format(ref, e)) from e
+
+            return pushed
