@@ -24,12 +24,12 @@ from contextlib import ExitStack
 from mmap import mmap
 
 import click
-import pkg_resources
 from ruamel import yaml
 
 from . import Profile
 from .. import Element, Scope, Consistency
 from .. import _yaml
+from .. import __version__ as bst_version
 from .._exceptions import ImplError
 from .._message import MessageType
 from ..plugin import _plugin_lookup
@@ -182,7 +182,7 @@ class ElementName(Widget):
             return ""
 
         plugin = _plugin_lookup(element_id)
-        name = plugin.name
+        name = plugin._get_full_name()
 
         # Sneak the action name in with the element name
         action_name = message.action_name
@@ -432,12 +432,11 @@ class LogLine(Widget):
         context = pipeline.context
         project = pipeline.project
         starttime = datetime.datetime.now()
-        bst = pkg_resources.require("buildstream")[0]
         text = ''
 
         # Main invocation context
         text += '\n'
-        text += self.content_profile.fmt("BuildStream Version {}\n".format(bst.version), bold=True)
+        text += self.content_profile.fmt("BuildStream Version {}\n".format(bst_version), bold=True)
         values = OrderedDict()
         values["Session Start"] = starttime.strftime('%A, %d-%m-%Y at %H:%M:%S')
         values["Project"] = "{} ({})".format(project.name, project.directory)
@@ -471,8 +470,8 @@ class LogLine(Widget):
             text += '\n'
 
         # Plugins
-        text += self.format_plugins(pipeline.element_factory.loaded_dependencies,
-                                    pipeline.source_factory.loaded_dependencies)
+        text += self.format_plugins(project._element_factory.loaded_dependencies,
+                                    project._source_factory.loaded_dependencies)
 
         # Pipeline state
         text += self.content_profile.fmt("Pipeline\n", bold=True)
@@ -585,7 +584,7 @@ class LogLine(Widget):
 
             full_key, cache_key, dim_keys = element._get_full_display_key()
 
-            line = p.fmt_subst(line, 'name', element.name, fg='blue', bold=True)
+            line = p.fmt_subst(line, 'name', element._get_full_name(), fg='blue', bold=True)
             line = p.fmt_subst(line, 'key', cache_key, fg='yellow', dim=dim_keys)
             line = p.fmt_subst(line, 'full-key', full_key, fg='yellow', dim=dim_keys)
 
