@@ -21,13 +21,18 @@
 import os
 from functools import cmp_to_key
 from collections import Mapping, namedtuple
+import tempfile
+import shutil
 
 from ._exceptions import LoadError, LoadErrorReason
+from . import Consistency
+from ._project import Project
 from . import _yaml
 
 from ._metaelement import MetaElement
 from ._metasource import MetaSource
 from ._profile import Topics, profile_start, profile_end
+from ._platform import Platform
 
 
 #################################################
@@ -51,15 +56,17 @@ class Symbol():
     RUNTIME = "runtime"
     ALL = "all"
     DIRECTORY = "directory"
+    JUNCTION = "junction"
 
 
 # A simple dependency object
 #
 class Dependency():
     def __init__(self, name,
-                 dep_type=None, provenance=None):
+                 dep_type=None, junction=None, provenance=None):
         self.name = name
         self.dep_type = dep_type
+        self.junction = junction
         self.provenance = provenance
 
 
@@ -69,11 +76,18 @@ class Dependency():
 #
 class LoadElement():
 
-    def __init__(self, data, filename, elements):
+    def __init__(self, data, filename, loader):
 
         self.data = data
         self.name = filename
-        self.elements = elements
+        self.loader = loader
+
+        if loader.project._junction:
+            # dependency is in subproject, qualify name
+            self.full_name = '{}:{}'.format(loader.project._junction.name, self.name)
+        else:
+            # dependency is in top-level project
+            self.full_name = self.name
 
         # Ensure the root node is valid
         _yaml.node_validate(self.data, [
@@ -98,7 +112,7 @@ class LoadElement():
     def depends(self, other):
 
         self.ensure_depends_cache()
-        return self.dep_cache.get(other.name) is not None
+        return self.dep_cache.get(other.full_name) is not None
 
     def ensure_depends_cache(self):
 
@@ -107,13 +121,13 @@ class LoadElement():
 
         self.dep_cache = {}
         for dep in self.deps:
-            elt = self.elements[dep.name]
+            elt = self.loader.get_element_for_dep(dep)
 
             # Ensure the cache of the element we depend on
             elt.ensure_depends_cache()
 
             # We depend on this element
-            self.dep_cache[dep.name] = True
+            self.dep_cache[elt.full_name] = True
 
             # And we depend on everything this element depends on
             self.dep_cache.update(elt.dep_cache)
@@ -137,7 +151,7 @@ def extract_depends_from_node(data):
             dependency = Dependency(dep, provenance=dep_provenance)
 
         elif isinstance(dep, Mapping):
-            _yaml.node_validate(dep, ['filename', 'type'])
+            _yaml.node_validate(dep, ['filename', 'type', 'junction'])
 
             # Make type optional, for this we set it to None after
             dep_type = _yaml.node_get(dep, str, Symbol.TYPE, default_value="")
@@ -150,8 +164,14 @@ def extract_depends_from_node(data):
                                 .format(provenance, dep_type))
 
             filename = _yaml.node_get(dep, str, Symbol.FILENAME)
+
+            junction = _yaml.node_get(dep, str, Symbol.JUNCTION, default_value="")
+            if not junction:
+                junction = None
+
             dependency = Dependency(filename,
-                                    dep_type=dep_type, provenance=dep_provenance)
+                                    dep_type=dep_type, junction=junction,
+                                    provenance=dep_provenance)
 
         else:
             index = depends.index(dep)
@@ -178,7 +198,9 @@ def extract_depends_from_node(data):
 #
 class Loader():
 
-    def __init__(self, basedir, filenames, options):
+    def __init__(self, project, filenames, *, parent=None, tempdir=None):
+
+        basedir = project.element_path
 
         # Ensure we have an absolute path for the base directory
         #
@@ -194,12 +216,21 @@ class Loader():
                                 "path to the base project directory: {}"
                                 .format(filename, basedir))
 
-        self.options = options   # Project options (OptionPool)
-        self.basedir = basedir   # Base project directory
-        self.targets = filenames   # Target bst elements
+        self.project = project
+        self.context = project._context
+        self.options = project._options     # Project options (OptionPool)
+        self.basedir = basedir              # Base project directory
+        self.targets = filenames            # Target bst elements
+        self.tempdir = tempdir
+
+        self.parent = parent
+
+        self.platform = Platform.get_platform()
+        self.artifacts = self.platform.artifactcache
 
         self.meta_elements = {}  # Dict of resolved meta elements by name
         self.elements = {}       # Dict of elements
+        self.loaders = {}        # Dict of junction loaders
 
     ########################################
     #           Main Entry Point           #
@@ -232,8 +263,8 @@ class Loader():
 
         # Set up a dummy element that depends on all top-level targets
         # to resolve potential circular dependencies between them
-        DummyTarget = namedtuple('DummyTarget', ['name', 'deps'])
-        dummy = DummyTarget(name='', deps=[Dependency(e) for e in self.targets])
+        DummyTarget = namedtuple('DummyTarget', ['name', 'full_name', 'deps'])
+        dummy = DummyTarget(name='', full_name='', deps=[Dependency(e) for e in self.targets])
         self.elements[''] = dummy
 
         profile_key = "_".join(t for t in self.targets)
@@ -253,6 +284,103 @@ class Loader():
         #
         return [self.collect_element(target) for target in self.targets]
 
+    # get_loader():
+    #
+    # Return loader for specified junction
+    #
+    # Args:
+    #    filename (str): Junction name
+    #
+    # Raises: LoadError
+    #
+    # Returns: A Loader or None if specified junction does not exist
+    def get_loader(self, filename, *, rewritable=False, ticker=None, level=0):
+        # return previously determined result
+        if filename in self.loaders:
+            loader = self.loaders[filename]
+
+            if loader is None:
+                # do not allow junctions with the same name in different
+                # subprojects
+                raise LoadError(LoadErrorReason.CONFLICTING_JUNCTION,
+                                "Conflicting junction {} in subprojects, define junction in {}"
+                                .format(filename, self.project.name))
+
+            return loader
+
+        if self.parent:
+            # junctions in the parent take precedence over junctions defined
+            # in subprojects
+            loader = self.parent.get_loader(filename, rewritable=rewritable, ticker=ticker, level=level + 1)
+            if loader:
+                self.loaders[filename] = loader
+                return loader
+
+        try:
+            load_element = self.load_file(filename, rewritable, ticker)
+        except LoadError as e:
+            if e.reason != LoadErrorReason.MISSING_FILE:
+                # other load error
+                raise
+
+            if level == 0:
+                # junction element not found in this or ancestor projects
+                raise
+            else:
+                # mark junction as not available to allow detection of
+                # conflicting junctions in subprojects
+                self.loaders[filename] = None
+                return None
+
+        # meta junction element
+        meta_element = self.collect_element(filename)
+
+        if meta_element.kind != 'junction':
+            raise LoadError(LoadErrorReason.INVALID_DATA,
+                            "{}: Expected junction but element kind is {}".format(filename, meta_element.kind))
+
+        element = meta_element.project._create_element(meta_element.kind,
+                                                       self.artifacts,
+                                                       meta_element)
+
+        os.makedirs(self.context.builddir, exist_ok=True)
+        basedir = tempfile.mkdtemp(prefix="{}-".format(element.normal_name), dir=self.context.builddir)
+
+        for meta_source in meta_element.sources:
+            source = meta_element.project._create_source(meta_source.kind,
+                                                         meta_source)
+
+            source.preflight()
+
+            if source.get_consistency() != Consistency.CACHED:
+                if self.context._fetch_subprojects:
+                    if ticker:
+                        ticker(filename, 'Fetching subproject from {} source'.format(meta_source.kind))
+                    source.fetch()
+                else:
+                    raise LoadError(LoadErrorReason.MISSING_FILE, "Subproject fetch needed for {}".format(filename))
+
+            source._stage(basedir)
+
+        project_dir = os.path.join(basedir, element.path)
+
+        try:
+            project = Project(project_dir, self.context, junction=element)
+        except LoadError as e:
+            if e.reason == LoadErrorReason.MISSING_FILE:
+                raise LoadError(reason=e.reason,
+                                message="Could not find the project.conf file for {}. "
+                                        "Expecting a project at path '{}' within {}".
+                                        format(element, element.path or '.', source)) from e
+            else:
+                raise
+
+        loader = Loader(project, [], parent=self, tempdir=basedir)
+
+        self.loaders[filename] = loader
+
+        return loader
+
     ########################################
     #             Loading Files            #
     ########################################
@@ -274,13 +402,24 @@ class Loader():
         data = _yaml.load(fullpath, shortname=filename, copy_tree=rewritable)
         self.options.process_node(data)
 
-        element = LoadElement(data, filename, self.elements)
+        element = LoadElement(data, filename, self)
 
         self.elements[filename] = element
 
         # Load all dependency files for the new LoadElement
         for dep in element.deps:
-            self.load_file(dep.name, rewritable, ticker)
+            if dep.junction:
+                self.load_file(dep.junction, rewritable, ticker)
+                loader = self.get_loader(dep.junction, rewritable=rewritable, ticker=ticker)
+            else:
+                loader = self
+
+            dep_element = loader.load_file(dep.name, rewritable, ticker)
+
+            if _yaml.node_get(dep_element.data, str, Symbol.KIND) == 'junction':
+                raise LoadError(LoadErrorReason.INVALID_DATA,
+                                "{}: Cannot depend on junction"
+                                .format(dep.provenance))
 
         return element
 
@@ -300,6 +439,10 @@ class Loader():
 
         element = self.elements[element_name]
 
+        # element name must be unique across projects
+        # to be usable as key for the check_elements and validated dicts
+        element_name = element.full_name
+
         # Skip already validated branches
         if validated.get(element_name) is not None:
             return
@@ -312,7 +455,8 @@ class Loader():
         # Push / Check each dependency / Pop
         check_elements[element_name] = True
         for dep in element.deps:
-            self.check_circular_deps(dep.name, check_elements, validated)
+            loader = self.get_loader_for_dep(dep)
+            loader.check_circular_deps(dep.name, check_elements, validated)
         del check_elements[element_name]
 
         # Eliminate duplicate paths
@@ -333,16 +477,22 @@ class Loader():
         if visited is None:
             visited = {}
 
+        element = self.elements[element_name]
+
+        # element name must be unique across projects
+        # to be usable as key for the visited dict
+        element_name = element.full_name
+
         if visited.get(element_name) is not None:
             return
 
-        element = self.elements[element_name]
         for dep in element.deps:
-            self.sort_dependencies(dep.name, visited=visited)
+            loader = self.get_loader_for_dep(dep)
+            loader.sort_dependencies(dep.name, visited=visited)
 
         def dependency_cmp(dep_a, dep_b):
-            element_a = self.elements[dep_a.name]
-            element_b = self.elements[dep_b.name]
+            element_a = self.get_element_for_dep(dep_a)
+            element_b = self.get_element_for_dep(dep_b)
 
             # Sort on inter element dependency first
             if element_a.depends(element_b):
@@ -364,6 +514,18 @@ class Loader():
             elif dep_a.name < dep_b.name:
                 return -1
 
+            # Sort local elements before junction elements
+            # and use string comparison between junction elements
+            if dep_a.junction and dep_b.junction:
+                if dep_a.junction > dep_b.junction:
+                    return 1
+                elif dep_a.junction < dep_b.junction:
+                    return -1
+            elif dep_a.junction:
+                return -1
+            elif dep_b.junction:
+                return 1
+
             # This wont ever happen
             return 0
 
@@ -419,7 +581,7 @@ class Loader():
             meta_sources.append(meta_source)
 
         kind = _yaml.node_get(data, str, Symbol.KIND)
-        meta_element = MetaElement(element_name, kind,
+        meta_element = MetaElement(self.project, element_name, kind,
                                    elt_provenance, meta_sources,
                                    _yaml.node_get(data, Mapping, Symbol.CONFIG, default_value={}),
                                    _yaml.node_get(data, Mapping, Symbol.VARIABLES, default_value={}),
@@ -432,10 +594,48 @@ class Loader():
 
         # Descend
         for dep in element.deps:
-            meta_dep = self.collect_element(dep.name)
+            if kind == 'junction':
+                raise LoadError(LoadErrorReason.INVALID_DATA,
+                                "{}: Junctions do not support dependencies".format(dep.provenance))
+
+            loader = self.get_loader_for_dep(dep)
+            meta_dep = loader.collect_element(dep.name)
             if dep.dep_type != 'runtime':
                 meta_element.build_dependencies.append(meta_dep)
             if dep.dep_type != 'build':
                 meta_element.dependencies.append(meta_dep)
 
         return meta_element
+
+    def get_loader_for_dep(self, dep):
+        if dep.junction:
+            # junction dependency, delegate to appropriate loader
+            return self.loaders[dep.junction]
+        else:
+            return self
+
+    def get_element_for_dep(self, dep):
+        return self.get_loader_for_dep(dep).elements[dep.name]
+
+    # cleanup():
+    #
+    # Remove temporary checkout directories of subprojects
+    def cleanup(self):
+        if self.parent and not self.tempdir:
+            # already done
+            return
+
+        # recurse
+        for loader in self.loaders.values():
+            # value may be None with nested junctions without overrides
+            if loader is not None:
+                loader.cleanup()
+
+        if not self.parent:
+            # basedir of top-level loader is never a temporary directory
+            return
+
+        # safe guard to not accidentally delete directories outside builddir
+        if self.tempdir.startswith(self.context.builddir + os.sep):
+            if os.path.exists(self.tempdir):
+                shutil.rmtree(self.tempdir)
