@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 #
-#  Copyright (C) 2016 Codethink Limited
+#  Copyright (C) 2016-2018 Codethink Limited
 #
 #  This program is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU Lesser General Public
@@ -27,19 +27,17 @@ import tarfile
 import itertools
 from contextlib import contextmanager
 from operator import itemgetter
-from pluginbase import PluginBase
 from tempfile import TemporaryDirectory
 
 from ._exceptions import PipelineError, ArtifactError, ImplError, BstError
 from ._message import Message, MessageType
-from ._elementfactory import ElementFactory
 from ._loader import Loader
-from ._sourcefactory import SourceFactory
 from . import Consistency
 from . import Scope
 from . import _site
 from . import utils
 from ._platform import Platform
+from ._artifactcache import ArtifactCacheSpec, configured_remote_artifact_cache_specs
 
 from ._scheduler import SchedStatus, TrackQueue, FetchQueue, BuildQueue, PullQueue, PushQueue
 
@@ -93,7 +91,13 @@ class Planner():
 #                         current source refs will not be the effective refs.
 #    rewritable (bool): Whether the loaded files should be rewritable
 #                       this is a bit more expensive due to deep copies
-#    use_remote_cache (bool): Whether to connect with remote artifact cache
+#    use_configured_remote_caches (bool): Whether to connect to configured artifact remotes.
+#    add_remote_cache (str): Adds an additional artifact remote URL, which is
+#                            prepended to the list of remotes (and thus given highest priority).
+#
+# The ticker methods will be called with an element name for each tick, a final
+# tick with None as the argument is passed to signal that processing of this
+# stage has terminated.
 #
 # Raises:
 #    LoadError
@@ -117,16 +121,10 @@ class Pipeline():
         self.platform = Platform.get_platform()
         self.artifacts = self.platform.artifactcache
 
-        loader = Loader(self.project.element_path, targets + except_,
-                        self.project._options)
+        self.loader = Loader(self.project, targets + except_)
 
         with self.timed_activity("Loading pipeline", silent_nested=True):
-            meta_elements = loader.load(rewritable, None)
-
-        # Create the factories after resolving the project
-        pluginbase = PluginBase(package='buildstream.plugins')
-        self.element_factory = ElementFactory(pluginbase, project._plugin_element_origins)
-        self.source_factory = SourceFactory(pluginbase, project._plugin_source_origins)
+            meta_elements = self.loader.load(rewritable, None)
 
         # Resolve the real elements now that we've resolved the project
         with self.timed_activity("Resolving pipeline"):
@@ -136,7 +134,8 @@ class Pipeline():
         self.targets = resolved_elements[:len(targets)]
         self.exceptions = resolved_elements[len(targets):]
 
-    def initialize(self, use_remote_cache=False, inconsistent=None):
+    def initialize(self, use_configured_remote_caches=False,
+                   add_remote_cache=None, track_elements=None):
         # Preflight directly, before ever interrogating caches or
         # anything.
         self.preflight()
@@ -145,10 +144,22 @@ class Pipeline():
 
         self.initialize_workspaces()
 
-        if use_remote_cache and self.artifacts.can_fetch():
-            self.fetch_remote_refs()
+        # Initialize remote artifact caches. We allow the commandline to override
+        # the user config in some cases (for example `bst push --remote=...`).
+        has_remote_caches = False
+        if add_remote_cache:
+            self.artifacts.set_remotes([ArtifactCacheSpec(add_remote_cache, push=True)])
+            has_remote_caches = True
+        if use_configured_remote_caches:
+            for project in self.context._get_projects():
+                artifact_caches = configured_remote_artifact_cache_specs(self.context, project)
+                if len(artifact_caches) > 0:
+                    self.artifacts.set_remotes(artifact_caches, project=project)
+                    has_remote_caches = True
+        if has_remote_caches:
+            self.initialize_remote_caches()
 
-        self.resolve_cache_keys(inconsistent)
+        self.resolve_cache_keys(track_elements)
 
     def preflight(self):
         for plugin in self.dependencies(Scope.ALL, include_sources=True):
@@ -164,39 +175,37 @@ class Pipeline():
                 raise PipelineError("{}: {}".format(plugin, e), reason=e.reason) from e
 
     def initialize_workspaces(self):
-        for element_name, source, workspace in self.project._list_workspaces():
+        for element_name, workspace in self.project._list_workspaces():
             for target in self.targets:
                 element = target.search(Scope.ALL, element_name)
 
                 if element is None:
-                    self.unused_workspaces.append((element_name, source, workspace))
+                    self.unused_workspaces.append((element_name, workspace))
                     continue
 
-                self.project._set_workspace(element, source, workspace)
+                self.project._set_workspace(element, workspace)
 
-    def fetch_remote_refs(self):
-        with self.timed_activity("Fetching remote refs", silent_nested=True):
-            try:
-                self.artifacts.initialize_remote()
-                self.artifacts.fetch_remote_refs()
-            except ArtifactError:
-                self.message(MessageType.WARN, "Failed to fetch remote refs")
-                self.artifacts.set_offline()
+    def initialize_remote_caches(self):
+        def remote_failed(url, error):
+            self.message(MessageType.WARN, "Failed to fetch remote refs from {}: {}".format(url, error))
 
-    def resolve_cache_keys(self, inconsistent):
-        if inconsistent:
-            inconsistent = self.get_elements_to_track(inconsistent)
+        with self.timed_activity("Initializing remote caches", silent_nested=True):
+            self.artifacts.initialize_remotes(on_failure=remote_failed)
+
+    def resolve_cache_keys(self, track_elements):
+        if track_elements:
+            track_elements = self.get_elements_to_track(track_elements)
 
         with self.timed_activity("Resolving cached state", silent_nested=True):
             for element in self.dependencies(Scope.ALL):
-                if inconsistent and element in inconsistent:
+                if track_elements and element in track_elements:
                     # Load the pipeline in an explicitly inconsistent state, use
                     # this for pipelines with tracking queues enabled.
-                    element._force_inconsistent()
-                else:
-                    # Resolve cache keys and interrogate the artifact cache
-                    # for the first time.
-                    element._cached()
+                    element._schedule_tracking()
+
+                # Determine initial element state. This may resolve cache keys
+                # and interrogate the artifact cache.
+                element._update_state()
 
     # Generator function to iterate over elements and optionally
     # also iterate over sources.
@@ -221,9 +230,10 @@ class Pipeline():
     #
     def assert_consistent(self, toplevel):
         inconsistent = []
-        for element in toplevel:
-            if element._consistency() == Consistency.INCONSISTENT:
-                inconsistent.append(element)
+        with self.timed_activity("Checking sources"):
+            for element in toplevel:
+                if element._consistency() == Consistency.INCONSISTENT:
+                    inconsistent.append(element)
 
         if inconsistent:
             detail = "Exact versions are missing for the following elements\n" + \
@@ -268,11 +278,9 @@ class Pipeline():
         if meta_element in self._resolved_elements:
             return self._resolved_elements[meta_element]
 
-        element = self.element_factory.create(meta_element.kind,
-                                              self.context,
-                                              self.project,
-                                              self.artifacts,
-                                              meta_element)
+        element = meta_element.project._create_element(meta_element.kind,
+                                                       self.artifacts,
+                                                       meta_element)
 
         self._resolved_elements[meta_element] = element
 
@@ -285,10 +293,8 @@ class Pipeline():
         # resolve sources
         for meta_source in meta_element.sources:
             element._add_source(
-                self.source_factory.create(meta_source.kind,
-                                           self.context,
-                                           self.project,
-                                           meta_source)
+                meta_element.project._create_source(meta_source.kind,
+                                                    meta_source)
             )
 
         return element
@@ -409,7 +415,7 @@ class Pipeline():
     def build(self, scheduler, build_all, track_first, save):
         if len(self.unused_workspaces) > 0:
             self.message(MessageType.WARN, "Unused workspaces",
-                         detail="\n".join([el + "-" + str(src) for el, src, _
+                         detail="\n".join([el for el, _
                                            in self.unused_workspaces]))
 
         # We set up two plans; one to track elements, the other to
@@ -446,12 +452,12 @@ class Pipeline():
         if track_plan:
             track = TrackQueue(save=save)
             queues.append(track)
-        if self.artifacts.can_fetch():
+        if self.artifacts.has_fetch_remotes():
             pull = PullQueue()
             queues.append(pull)
         queues.append(fetch)
         queues.append(build)
-        if self.artifacts.can_push():
+        if self.artifacts.has_push_remotes():
             push = PushQueue()
             queues.append(push)
 
@@ -503,18 +509,22 @@ class Pipeline():
                                 .format(directory))
 
         # Stage deps into a temporary sandbox first
-        with target._prepare_sandbox(Scope.RUN, None, integrate=integrate) as sandbox:
+        try:
+            with target._prepare_sandbox(Scope.RUN, None, integrate=integrate) as sandbox:
 
-            # Copy or move the sandbox to the target directory
-            sandbox_root = sandbox.get_directory()
-            with target.timed_activity("Checking out files in {}".format(directory)):
-                try:
-                    if hardlinks:
-                        self.checkout_hardlinks(sandbox_root, directory)
-                    else:
-                        utils.copy_files(sandbox_root, directory)
-                except OSError as e:
-                    raise PipelineError("Failed to checkout files: {}".format(e)) from e
+                # Copy or move the sandbox to the target directory
+                sandbox_root = sandbox.get_directory()
+                with target.timed_activity("Checking out files in {}".format(directory)):
+                    try:
+                        if hardlinks:
+                            self.checkout_hardlinks(sandbox_root, directory)
+                        else:
+                            utils.copy_files(sandbox_root, directory)
+                    except OSError as e:
+                        raise PipelineError("Failed to checkout files: {}".format(e)) from e
+        except BstError as e:
+            raise PipelineError("Error while staging dependencies into a sandbox: {}".format(e),
+                                reason=e.reason) from e
 
     # Helper function for checkout()
     #
@@ -541,17 +551,17 @@ class Pipeline():
     #
     # Args:
     #    directory (str): The directory to stage the source in
-    #    source_index (int): The index of the source to stage
     #    no_checkout (bool): Whether to skip checking out the source
     #    track_first (bool): Whether to track and fetch first
     #    force (bool): Whether to ignore contents in an existing directory
     #
-    def open_workspace(self, scheduler, directory, source_index, no_checkout, track_first, force):
+    def open_workspace(self, scheduler, directory, no_checkout, track_first, force):
         # When working on workspaces we only have one target
         target = self.targets[0]
         workdir = os.path.abspath(directory)
-        sources = list(target.sources())
-        source_index = self.validate_workspace_index(source_index)
+
+        if len(list(target.sources())) == 0:
+            raise PipelineError("The given element has no sources")
 
         # Check directory
         try:
@@ -563,9 +573,9 @@ class Pipeline():
             raise PipelineError("Checkout directory is not empty: {}".format(directory))
 
         # Check for workspace config
-        if self.project._get_workspace(target.name, source_index):
+        if self.project._get_workspace(target.name):
             raise PipelineError("Workspace '{}' is already defined."
-                                .format(target.name + " - " + str(source_index)))
+                                .format(target.name))
 
         plan = [target]
 
@@ -599,16 +609,16 @@ class Pipeline():
                              "Fetched {} elements".format(fetched), elapsed=elapsed)
 
         if not no_checkout:
-            source = sources[source_index]
-            with target.timed_activity("Staging source to {}".format(directory)):
-                if source.get_consistency() != Consistency.CACHED:
-                    raise PipelineError("Could not stage uncached source. " +
-                                        "Use `--track` to track and " +
-                                        "fetch the latest version of the " +
-                                        "source.")
-                source._init_workspace(directory)
+            with target.timed_activity("Staging sources to {}".format(directory)):
+                for source in target.sources():
+                    if source.get_consistency() != Consistency.CACHED:
+                        raise PipelineError("Could not stage uncached source. " +
+                                            "Use `--track` to track and " +
+                                            "fetch the latest version of the " +
+                                            "source.")
+                    source._init_workspace(directory)
 
-        self.project._set_workspace(target, source_index, workdir)
+        self.project._set_workspace(target, workdir)
 
         with target.timed_activity("Saving workspace configuration"):
             self.project._save_workspace_config()
@@ -618,17 +628,15 @@ class Pipeline():
     # Close a project workspace
     #
     # Args:
-    #    source_index (int) - The index of the source
     #    remove_dir (bool) - Whether to remove the associated directory
     #
-    def close_workspace(self, source_index, remove_dir):
+    def close_workspace(self, remove_dir):
         # When working on workspaces we only have one target
         target = self.targets[0]
-        source_index = self.validate_workspace_index(source_index)
 
         # Remove workspace directory if prompted
         if remove_dir:
-            path = self.project._get_workspace(target.name, source_index)
+            path = self.project._get_workspace(target.name)
             if path is not None:
                 with target.timed_activity("Removing workspace directory {}"
                                            .format(path)):
@@ -641,17 +649,17 @@ class Pipeline():
         # Delete the workspace config entry
         with target.timed_activity("Removing workspace"):
             try:
-                self.project._delete_workspace(target.name, source_index)
+                self.project._delete_workspace(target.name)
             except KeyError:
                 raise PipelineError("Workspace '{}' is currently not defined"
-                                    .format(target.name + " - " + str(source_index)))
+                                    .format(target.name))
 
         # Update workspace config
         self.project._save_workspace_config()
 
         # Reset source to avoid checking out the (now empty) workspace
-        source = list(target.sources())[source_index]
-        source._del_workspace()
+        for source in target.sources():
+            source._del_workspace()
 
     # reset_workspace
     #
@@ -664,20 +672,18 @@ class Pipeline():
     #    track (bool): Whether to also track the source
     #    no_checkout (bool): Whether to check out the source (at all)
     #
-    def reset_workspace(self, scheduler, source_index, track, no_checkout):
+    def reset_workspace(self, scheduler, track, no_checkout):
         # When working on workspaces we only have one target
         target = self.targets[0]
-        source_index = self.validate_workspace_index(source_index)
-        workspace_dir = self.project._get_workspace(target.name, source_index)
+        workspace_dir = self.project._get_workspace(target.name)
 
         if workspace_dir is None:
             raise PipelineError("Workspace '{}' is currently not defined"
                                 .format(target.name + " - " + str(source_index)))
 
-        self.close_workspace(source_index, True)
+        self.close_workspace(True)
 
-        self.open_workspace(scheduler, workspace_dir, source_index, no_checkout,
-                            track, False)
+        self.open_workspace(scheduler, workspace_dir, no_checkout, track, False)
 
     # pull()
     #
@@ -689,8 +695,8 @@ class Pipeline():
     #
     def pull(self, scheduler, elements):
 
-        if not self.artifacts.can_fetch():
-            raise PipelineError("Not configured for pulling artifacts")
+        if not self.artifacts.has_fetch_remotes():
+            raise PipelineError("Not artifact caches available for pulling artifacts")
 
         plan = elements
         self.assert_consistent(plan)
@@ -727,8 +733,8 @@ class Pipeline():
     #
     def push(self, scheduler, elements):
 
-        if not self.artifacts.can_push():
-            raise PipelineError("Not configured for pushing artifacts")
+        if not self.artifacts.has_push_remotes():
+            raise PipelineError("No artifact caches available for pushing artifacts")
 
         plan = elements
         self.assert_consistent(plan)
@@ -814,19 +820,6 @@ class Pipeline():
         # in before.
         return [element for element in elements if element in visited]
 
-    def validate_workspace_index(self, source_index):
-        sources = list(self.targets[0].sources())
-
-        # Validate source_index
-        if len(sources) < 1:
-            raise PipelineError("The given element has no sources")
-        if len(sources) == 1 and source_index is None:
-            source_index = 0
-        if source_index is None:
-            raise PipelineError("An index needs to be specified for elements with more than one source")
-
-        return source_index
-
     # Various commands define a --deps option to specify what elements to
     # use in the result, this function reports a list that is appropriate for
     # the selected option.
@@ -935,7 +928,7 @@ class Pipeline():
         with open(_site.build_all_template, "r") as f:
             script_template = f.read()
 
-        with open(script_path, "w") as script:
+        with utils.save_file_atomic(script_path, "w") as script:
             script.write(script_template.format(modules=module_string))
 
         os.chmod(script_path, stat.S_IEXEC | stat.S_IREAD)
@@ -950,3 +943,7 @@ class Pipeline():
 
             with tarfile.open(tar_name, permissions) as tar:
                 tar.add(directory, arcname=element_name)
+
+    def cleanup(self):
+        if self.loader:
+            self.loader.cleanup()
