@@ -19,7 +19,7 @@
 #        Tristan Van Berkom <tristan.vanberkom@codethink.co.uk>
 #        Jürg Billeter <juerg.billeter@codethink.co.uk>
 
-from . import Queue, QueueType
+from . import Queue, QueueStatus, QueueType
 
 
 # A queue which assembles elements
@@ -30,19 +30,30 @@ class BuildQueue(Queue):
     complete_name = "Built"
     queue_type = QueueType.BUILD
 
+    def prepare(self, element):
+        # Inform element in main process that it is scheduled for assembly
+        element._schedule_assemble()
+
     def process(self, element):
         element._assemble()
         return element._get_unique_id()
 
-    def ready(self, element):
-        return element._buildable()
+    def status(self, element):
+        # state of dependencies may have changed, recalculate element state
+        element._update_state()
 
-    def skip(self, element):
-        return element._cached()
+        if element._cached():
+            return QueueStatus.SKIP
+
+        if not element._buildable():
+            return QueueStatus.WAIT
+
+        return QueueStatus.READY
 
     def done(self, element, result, returncode):
-        # Elements are cached after they are successfully assembled
-        if returncode == 0:
-            element._cached(recalculate=True)
+        # Inform element in main process that assembly is done
+        element._assemble_done()
+
+        element._update_state()
 
         return True
