@@ -27,7 +27,7 @@ from ..plugin import _plugin_lookup
 from .. import _yaml
 
 # Local imports
-from . import Queue, QueueType
+from . import Queue, QueueStatus, QueueType
 
 
 # A queue which tracks sources
@@ -45,9 +45,12 @@ class TrackQueue(Queue):
     def process(self, element):
         return element._track()
 
-    def skip(self, element):
+    def status(self, element):
         # We can skip elements entirely if they have no sources.
-        return len(list(element.sources())) == 0
+        if len(list(element.sources())) == 0:
+            return QueueStatus.SKIP
+
+        return QueueStatus.READY
 
     def done(self, element, result, returncode):
 
@@ -83,13 +86,9 @@ class TrackQueue(Queue):
                                     "tracked source to file {}: {}"
                                     .format(source, fullname, e))
 
-        # Forcefully recalculate the element's consistency state after successfully
-        # tracking, this is avoid a following fetch queue operating on the sources
-        # if the tracked ref is cached as a result.
-        #
         context = element._get_context()
         context._push_message_depth(True)
-        element._consistency(recalculate=True)
+        element._update_state()
         context._pop_message_depth()
 
         # We'll appear as a skipped element if tracking resulted in no change
