@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 #
-#  Copyright (C) 2016 Codethink Limited
+#  Copyright (C) 2016-2018 Codethink Limited
 #
 #  This program is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU Lesser General Public
@@ -37,14 +37,24 @@ from . import _yaml
 from ._variables import Variables
 from ._exceptions import BstError, LoadError, LoadErrorReason, ImplError, ErrorDomain
 from . import Plugin, Consistency
-from ._project import BST_ARTIFACT_VERSION as BST_CORE_ARTIFACT_VERSION
 from . import SandboxFlags
 from . import utils
+from . import _cachekey
 from . import _signals
 from . import _site
 from ._platform import Platform
 
 
+# The base BuildStream artifact version
+#
+# The artifact version changes whenever the cache key
+# calculation algorithm changes in an incompatible way
+# or if buildstream was changed in a way which can cause
+# the same cache key to produce something that is no longer
+# the same.
+_BST_CORE_ARTIFACT_VERSION = 1
+
+
 # _KeyStrength():
 #
 # Strength of cache key
@@ -85,10 +95,11 @@ class ElementError(BstError):
 
     Args:
        message (str): The error message to report to the user
+       detail (str): A possibly multiline, more detailed error message
        reason (str): An optional machine readable reason string, used for test cases
     """
-    def __init__(self, message, *, reason=None):
-        super().__init__(message, domain=ErrorDomain.ELEMENT, reason=reason)
+    def __init__(self, message, *, detail=None, reason=None):
+        super().__init__(message, detail=detail, domain=ErrorDomain.ELEMENT, reason=reason)
 
 
 class Element(Plugin):
@@ -132,14 +143,19 @@ class Element(Plugin):
         self.__runtime_dependencies = []        # Direct runtime dependency Elements
         self.__build_dependencies = []          # Direct build dependency Elements
         self.__sources = []                     # List of Sources
+        self.__cache_key_dict = None            # Dict for cache key calculation
         self.__cache_key = None                 # Our cached cache key
         self.__weak_cache_key = None            # Our cached weak cache key
-        self.__cache_key_from_artifact = None   # Our cached cache key from artifact
+        self.__strict_cache_key = None          # Our cached cache key for strict builds
         self.__artifacts = artifacts            # Artifact cache
         self.__cached = None                    # Whether we have a cached artifact
+        self.__strong_cached = None             # Whether we have a cached artifact
         self.__remotely_cached = None           # Whether we have a remotely cached artifact
+        self.__remotely_strong_cached = None    # Whether we have a remotely cached artifact
+        self.__pull_failed = False              # Whether pull was attempted but failed
         self.__log_path = None                  # Path to dedicated log file or None
         self.__splits = None
+        self.__whitelist_regex = None
 
         # Ensure we have loaded this class's defaults
         self.__init_defaults(plugin_conf)
@@ -202,17 +218,19 @@ class Element(Plugin):
         if visited is None:
             visited = {}
 
+        full_name = self._get_full_name()
+
         scope_set = set((Scope.BUILD, Scope.RUN)) if scope == Scope.ALL else set((scope,))
 
-        if self.name in visited and scope_set.issubset(visited[self.name]):
+        if full_name in visited and scope_set.issubset(visited[full_name]):
             return
 
         should_yield = False
-        if self.name not in visited:
-            visited[self.name] = scope_set
+        if full_name not in visited:
+            visited[full_name] = scope_set
             should_yield = True
         else:
-            visited[self.name] |= scope_set
+            visited[full_name] |= scope_set
 
         if recurse or not recursed:
             if scope == Scope.ALL:
@@ -437,7 +455,8 @@ class Element(Plugin):
 
         Raises:
            (:class:`.ElementError`): If any of the dependencies in `scope` have not
-                                     yet produced artifacts.
+                                     yet produced artifacts, or if forbidden overlaps
+                                     occur.
         """
         ignored = {}
         overlaps = OrderedDict()
@@ -466,10 +485,32 @@ class Element(Plugin):
                 ignored[dep.name] = result.ignored
 
         if overlaps:
-            detail = "Staged files overwrite existing files in staging area:\n"
+            overlap_error = overlap_warning = False
+            error_detail = warning_detail = "Staged files overwrite existing files in staging area:\n"
             for f, elements in overlaps.items():
-                detail += "  /{}: ".format(f) + " above ".join(reversed(elements)) + "\n"
-            self.warn("Overlaps detected", detail=detail)
+                overlap_error_elements = []
+                overlap_warning_elements = []
+                # The bottom item overlaps nothing
+                overlapping_elements = elements[1:]
+                for elm in overlapping_elements:
+                    element = self.search(scope, elm)
+                    element_project = element._get_project()
+                    if not element.__file_is_whitelisted(f):
+                        if element_project._fail_on_overlap:
+                            overlap_error_elements.append(elm)
+                            overlap_error = True
+                        else:
+                            overlap_warning_elements.append(elm)
+                            overlap_warning = True
+
+                warning_detail += _overlap_error_detail(f, overlap_warning_elements, elements)
+                error_detail += _overlap_error_detail(f, overlap_error_elements, elements)
+
+            if overlap_warning:
+                self.warn("Non-whitelisted overlaps detected", detail=warning_detail)
+            if overlap_error:
+                raise ElementError("Non-whitelisted overlaps detected and fail-on-overlaps is set",
+                                   detail=error_detail, reason="overlap-error")
 
         if ignored:
             detail = "Not staging files which would replace non-empty directories:\n"
@@ -510,10 +551,7 @@ class Element(Plugin):
            directory (str): An absolute path within the sandbox to stage the sources at
         """
 
-        sandbox_root = sandbox.get_directory()
-        host_directory = os.path.join(sandbox_root, directory.lstrip(os.sep))
-
-        self._stage_sources_at(host_directory)
+        self._stage_sources_in_sandbox(sandbox, directory)
 
     def get_public_data(self, domain):
         """Fetch public data on this element
@@ -688,7 +726,7 @@ class Element(Plugin):
         script_path = os.path.join(directory, "build-" + self.normal_name)
 
         with self.timed_activity("Writing build script", silent_nested=True):
-            with open(script_path, "w") as script_file:
+            with utils.save_file_atomic(script_path, "w") as script_file:
                 script_file.write(script)
 
             os.chmod(script_path, stat.S_IEXEC | stat.S_IREAD)
@@ -712,27 +750,24 @@ class Element(Plugin):
 
     # _consistency():
     #
-    # Args:
-    #    recalcualte (bool): Whether to forcefully recalculate
-    #
     # Returns:
     #    (list): The minimum consistency of the elements sources
     #
     # If the element has no sources, this returns Consistency.CACHED
-    def _consistency(self, recalculate=False):
+    def _consistency(self):
 
         # The source objects already cache the consistency state, it
         # should not be expensive to iterate over the sources to get at it
         consistency = Consistency.CACHED
         for source in self.__sources:
-            source_consistency = source._get_consistency(recalculate=recalculate)
+            source_consistency = source._get_consistency()
             consistency = min(consistency, source_consistency)
         return consistency
 
-    # _force_inconsistent():
+    # _schedule_tracking():
     #
-    # Force an element state to be inconsistent. Cache keys are unset, Artifacts appear
-    # to be not cached and any sources appear to be inconsistent.
+    # Force an element state to be inconsistent. Any sources appear to be
+    # inconsistent.
     #
     # This is used across the pipeline in sessions where the
     # elements in question are going to be tracked, causing the
@@ -740,42 +775,45 @@ class Element(Plugin):
     # and reinterrogation of element state after tracking of elements
     # succeeds.
     #
-    def _force_inconsistent(self):
-        self.__cached = None
-        self.__cache_key = None
-        self.__weak_cache_key = None
+    def _schedule_tracking(self):
         for source in self.__sources:
-            source._force_inconsistent()
+            source._schedule_tracking()
+
+        self._update_state()
+
+    # _schedule_assemble():
+    #
+    # This is called in the main process before the element is assembled
+    # in a subprocess.
+    #
+    def _schedule_assemble(self):
+        for source in self.__sources:
+            source._schedule_assemble()
+
+        self._update_state()
+
+    # _assemble_done():
+    #
+    # This is called in the main process after the element has been assembled
+    # in a subprocess.
+    #
+    def _assemble_done(self):
+        for source in self.__sources:
+            source._assemble_done()
+
+        self._update_state()
 
     # _cached():
     #
-    # Args:
-    #    recalculate (bool): Whether to forcefully recalculate
-    #
     # Returns:
     #    (bool): Whether this element is already present in
     #            the artifact cache
     #
-    # Note: The recalculate argument is actually tristate:
-    #
-    #    o None: Calculate cache state if not previously calculated
-    #    o True: Force recalculate cached state, even if already checked
-    #    o False: Only return cached state, never recalculate automatically
-    #
-    def _cached(self, recalculate=None, strength=None):
-
-        if recalculate:
-            self.__cached = None
-            self.__strong_cached = None
+    def _cached(self, strength=None):
 
         if strength is None:
             strength = _KeyStrength.STRONG if self._get_strict() else _KeyStrength.WEAK
 
-        if recalculate is not False:
-            if self.__cached is None and self._get_cache_key() is not None:
-                self.__cached = self.__artifacts.contains(self)
-                self.__strong_cached = self.__artifacts.contains(self, strength=_KeyStrength.STRONG)
-
         if self.__cached is None:
             return False
         elif strength == _KeyStrength.STRONG:
@@ -785,45 +823,24 @@ class Element(Plugin):
 
     # _assert_cached()
     #
-    # Args:
-    #    recalculate (bool): Argument to pass to Element._cached()
-    #
     # Raises an error if the artifact is not cached.
     #
-    def _assert_cached(self, recalculate=None):
-        if not self._cached(recalculate=recalculate):
+    def _assert_cached(self):
+        if not self._cached():
             raise ElementError("{}: Missing artifact {}"
                                .format(self, self._get_display_key()))
 
     # _remotely_cached():
     #
-    # Args:
-    #    recalculate (bool): Whether to forcefully recalculate
-    #
     # Returns:
     #    (bool): Whether this element is already present in
     #            the remote artifact cache
     #
-    # Note: The recalculate argument is actually tristate:
-    #
-    #    o None: Calculate cache state if not previously calculated
-    #    o True: Force recalculate cached state, even if already checked
-    #    o False: Only return cached state, never recalculate automatically
-    #
-    def _remotely_cached(self, recalculate=None, strength=None):
-
-        if recalculate:
-            self.__remotely_cached = None
-            self.__remotely_strong_cached = None
+    def _remotely_cached(self, strength=None):
 
         if strength is None:
             strength = _KeyStrength.STRONG if self._get_strict() else _KeyStrength.WEAK
 
-        if recalculate is not False:
-            if self.__remotely_cached is None and self._get_cache_key() is not None:
-                self.__remotely_cached = self.__artifacts.remote_contains(self)
-                self.__remotely_strong_cached = self.__artifacts.remote_contains(self, strength=_KeyStrength.STRONG)
-
         if self.__remotely_cached is None:
             return False
         elif strength == _KeyStrength.STRONG:
@@ -891,31 +908,47 @@ class Element(Plugin):
         if None in dependencies:
             return None
 
-        # Filter out nocache variables from the element's environment
-        cache_env = {
-            key: value
-            for key, value in self.node_items(self.__environment)
-            if key not in self.__env_nocache
-        }
+        # Generate dict that is used as base for all cache keys
+        if self.__cache_key_dict is None:
+            # Filter out nocache variables from the element's environment
+            cache_env = {
+                key: value
+                for key, value in self.node_items(self.__environment)
+                if key not in self.__env_nocache
+            }
 
-        context = self._get_context()
-        project = self._get_project()
-        return utils._generate_key({
-            'artifact-version': "{}.{}".format(BST_CORE_ARTIFACT_VERSION,
-                                               self.BST_ARTIFACT_VERSION),
-            'context': context._get_cache_key(),
-            'project': project._get_cache_key(),
-            'element': self.get_unique_key(),
-            'environment': cache_env,
-            'sources': [s._get_unique_key() for s in self.__sources],
-            'dependencies': dependencies,
-            'public': self.__public,
-            'cache': type(self.__artifacts).__name__
-        })
+            context = self._get_context()
+            project = self._get_project()
+            operating_system, _, _, _, machine_arch = os.uname()
+
+            self.__cache_key_dict = {
+                'artifact-version': "{}.{}".format(_BST_CORE_ARTIFACT_VERSION,
+                                                   self.BST_ARTIFACT_VERSION),
+                'context': context._get_cache_key(),
+                'project': project._get_cache_key(),
+                'element': self.get_unique_key(),
+
+                # The execution environment may later be delegated
+                # to sandboxes which support virtualization
+                #
+                'execution-environment': {
+                    'os': operating_system,
+                    'arch': machine_arch
+                },
+                'environment': cache_env,
+                'sources': [s._get_unique_key() for s in self.__sources],
+                'public': self.__public,
+                'cache': type(self.__artifacts).__name__
+            }
+
+        cache_key_dict = self.__cache_key_dict.copy()
+        cache_key_dict['dependencies'] = dependencies
+
+        return _cachekey.generate_key(cache_key_dict)
 
     # _get_cache_key():
     #
-    # Returns the cache key, calculating it if necessary
+    # Returns the cache key
     #
     # Args:
     #    strength (_KeyStrength): Either STRONG or WEAK key strength
@@ -926,79 +959,22 @@ class Element(Plugin):
     # None is returned if information for the cache key is missing.
     #
     def _get_cache_key(self, strength=_KeyStrength.STRONG):
-        if self.__cache_key is None:
-            # It is not really necessary to check if the Source object's
-            # local mirror has the ref cached locally or not, it's only important
-            # to know if the source has a ref specified or not, in order to
-            # produce a cache key.
-            #
-            if self._consistency() == Consistency.INCONSISTENT:
-                return None
-
-            # Calculate strong cache key
-            dependencies = [
-                e._get_cache_key() for e in self.dependencies(Scope.BUILD)
-            ]
-            self.__cache_key = self.__calculate_cache_key(dependencies)
-
-            # Calculate weak cache key
-            # Weak cache key includes names of direct build dependencies
-            # but does not include keys of dependencies.
-            if self.BST_STRICT_REBUILD:
-                dependencies = [
-                    e._get_cache_key(strength=_KeyStrength.WEAK)
-                    for e in self.dependencies(Scope.BUILD)
-                ]
-            else:
-                dependencies = [
-                    e.name for e in self.dependencies(Scope.BUILD, recurse=False)
-                ]
-
-            self.__weak_cache_key = self.__calculate_cache_key(dependencies)
-
         if strength == _KeyStrength.STRONG:
             return self.__cache_key
         else:
             return self.__weak_cache_key
 
-    # _get_cache_key_from_artifact():
+    # _get_strict_cache_key():
     #
-    # Returns the strong cache key as stored in the cached artifact
-    #
-    # Args:
-    #    recalculate (bool): Whether to forcefully recalculate
+    # Returns the cache key for strict build mode
     #
     # Returns:
-    #    (str): A hex digest cache key for this Element
+    #    (str): A hex digest cache key for this Element, or None
     #
-    def _get_cache_key_from_artifact(self, recalculate=False):
-        if recalculate:
-            self.__cache_key_from_artifact = None
-
-        if self.__cache_key_from_artifact is None:
-            self._assert_cached(recalculate=False)
-
-            # Load the strong cache key from the artifact
-            metadir = os.path.join(self.__artifacts.extract(self), 'meta')
-            meta = _yaml.load(os.path.join(metadir, 'artifact.yaml'))
-            self.__cache_key_from_artifact = meta['keys']['strong']
-
-        return self.__cache_key_from_artifact
-
-    # _get_cache_key_for_build():
+    # None is returned if information for the cache key is missing.
     #
-    # Returns the strong cache key using cached artifacts as dependencies
-    #
-    # Returns:
-    #    (str): A hex digest cache key for this Element
-    #
-    # This is the cache key for a fresh build of this element.
-    #
-    def _get_cache_key_for_build(self):
-        dependencies = [
-            e._get_cache_key_from_artifact() for e in self.dependencies(Scope.BUILD)
-        ]
-        return self.__calculate_cache_key(dependencies)
+    def _get_strict_cache_key(self):
+        return self.__strict_cache_key
 
     # _get_full_display_key():
     #
@@ -1013,21 +989,15 @@ class Element(Plugin):
     #
     def _get_full_display_key(self):
         context = self._get_context()
-        cache_key = None
         dim_key = True
 
-        if self._consistency() == Consistency.INCONSISTENT:
-            cache_key = None
-        elif self._get_strict() or self._cached(strength=_KeyStrength.STRONG):
-            cache_key = self._get_cache_key()
-        elif self._cached():
-            cache_key = self._get_cache_key_from_artifact()
-        elif self._buildable():
-            cache_key = self._get_cache_key_for_build()
+        cache_key = self._get_cache_key()
 
         if not cache_key:
             cache_key = "{:?<64}".format('')
-        elif self._get_cache_key() == cache_key:
+        elif self._get_cache_key() == self.__strict_cache_key:
+            # Strong cache key used in this session matches cache key
+            # that would be used in strict build mode
             dim_key = False
 
         length = min(len(cache_key), context.log_key_length)
@@ -1064,13 +1034,12 @@ class Element(Plugin):
     #    (list): A list of Source object ids and their new references
     #
     def _track(self):
-        changed = []
+        refs = []
         for source in self.__sources:
             new_ref = source._track()
-            if new_ref is not None:
-                changed.append((source._get_unique_id(), new_ref))
+            refs.append((source._get_unique_id(), new_ref))
 
-        return changed
+        return refs
 
     # _assemble():
     #
@@ -1145,16 +1114,19 @@ class Element(Plugin):
                 # Store public data
                 _yaml.dump(_yaml.node_sanitize(self.__dynamic_public), os.path.join(metadir, 'public.yaml'))
 
+                # ensure we have cache keys
+                self._assemble_done()
+
                 # Store artifact metadata
                 dependencies = {
-                    e.name: e._get_cache_key_from_artifact() for e in self.dependencies(Scope.BUILD)
+                    e.name: e._get_cache_key() for e in self.dependencies(Scope.BUILD)
                 }
                 workspaced_dependencies = {
                     e.name: e._workspaced() for e in self.dependencies(Scope.BUILD)
                 }
                 meta = {
                     'keys': {
-                        'strong': self._get_cache_key_for_build(),
+                        'strong': self._get_cache_key(),
                         'weak': self._get_cache_key(_KeyStrength.WEAK),
                         'dependencies': dependencies
                     },
@@ -1195,15 +1167,18 @@ class Element(Plugin):
     #   (bool): True if this element should not be pushed
     #
     def _skip_push(self):
-        if not self._cached(recalculate=None):
+        if not self.__artifacts.has_push_remotes(element=self):
+            # No push remotes for this element's project
+            return True
+
+        if not self._cached():
             return True
 
         # Do not push tained artifact
         if self._tainted():
             return True
 
-        # Do not push artifact that is already in the remote artifact repository
-        return self.__artifacts.remote_contains_key(self, self._get_cache_key_from_artifact())
+        return False
 
     # _push():
     #
@@ -1262,8 +1237,9 @@ class Element(Plugin):
 
     # Set a source's workspace
     #
-    def _set_source_workspace(self, source_index, path):
-        self.__sources[source_index]._set_workspace(path)
+    def _set_source_workspaces(self, path):
+        for source in self.sources():
+            source._set_workspace(path)
 
     # Whether this element has a source that is workspaced.
     #
@@ -1287,7 +1263,7 @@ class Element(Plugin):
     def _workspaced_artifact(self):
 
         if self.__workspaced_artifact is None:
-            self._assert_cached(recalculate=False)
+            self._assert_cached()
 
             metadir = os.path.join(self.__artifacts.extract(self), 'meta')
             meta = _yaml.load(os.path.join(metadir, 'artifact.yaml'))
@@ -1301,7 +1277,7 @@ class Element(Plugin):
     def _workspaced_dependencies_artifact(self):
 
         if self.__workspaced_dependencies_artifact is None:
-            self._assert_cached(recalculate=False)
+            self._assert_cached()
 
             metadir = os.path.join(self.__artifacts.extract(self), 'meta')
             meta = _yaml.load(os.path.join(metadir, 'artifact.yaml'))
@@ -1400,19 +1376,26 @@ class Element(Plugin):
     def _shell(self, scope=None, directory=None, command=None):
 
         with self._prepare_sandbox(scope, directory) as sandbox:
-
-            # Override the element environment with some of
-            # the host environment and use that for the shell environment.
-            #
-            # XXX Hard code should be removed
             environment = self.get_environment()
-            environment = copy.copy(environment)
-            overrides = ['DISPLAY', 'DBUS_SESSION_BUS_ADDRESS']
-            for override in overrides:
-                if os.environ.get(override) is not None:
-                    environment[override] = os.environ.get(override)
+            flags = SandboxFlags.INTERACTIVE
 
-            flags = SandboxFlags.NETWORK_ENABLED | SandboxFlags.INTERACTIVE
+            if scope == Scope.RUN:
+                # If a testing sandbox was requested, override the element environment
+                # with some of the host environment and use that for the shell.
+                #
+                # XXX Hard code should be removed
+                environment = copy.copy(environment)
+                overrides = ['DISPLAY', 'DBUS_SESSION_BUS_ADDRESS']
+                for override in overrides:
+                    if os.environ.get(override) is not None:
+                        environment[override] = os.environ.get(override)
+
+            # Decide whether to create a build style sandbox or an open
+            # testing sandbox based on whether the build scope was requested
+            if scope == Scope.BUILD:
+                flags |= SandboxFlags.ROOT_READ_ONLY
+            elif scope == Scope.RUN:
+                flags |= SandboxFlags.NETWORK_ENABLED
 
             if command:
                 argv = [arg for arg in command]
@@ -1424,20 +1407,53 @@ class Element(Plugin):
             # Run shells with network enabled and readonly root.
             return sandbox.run(argv, flags, env=environment)
 
+    # _stage_sources_in_sandbox():
+    #
+    # Stage this element's sources to a directory inside sandbox
+    #
+    # Args:
+    #     sandbox (:class:`.Sandbox`): The build sandbox
+    #     directory (str): An absolute path to stage the sources at
+    #     mount_workspaces (bool): mount workspaces if True, copy otherwise
+    #
+    def _stage_sources_in_sandbox(self, sandbox, directory, mount_workspaces=True):
+
+        if mount_workspaces:
+            # First, mount sources that have an open workspace
+            sources_to_mount = [source for source in self.sources() if source._has_workspace()]
+            for source in sources_to_mount:
+                mount_point = source._get_staging_path(directory)
+                mount_source = source._get_workspace_path()
+                sandbox.mark_directory(mount_point)
+                sandbox._set_mount_source(mount_point, mount_source)
+
+        # Stage all sources that need to be copied
+        sandbox_root = sandbox.get_directory()
+        host_directory = os.path.join(sandbox_root, directory.lstrip(os.sep))
+        self._stage_sources_at(host_directory, mount_workspaces=mount_workspaces)
+
     # _stage_sources_at():
     #
     # Stage this element's sources to a directory
     #
     # Args:
     #     directory (str): An absolute path to stage the sources at
+    #     mount_workspaces (bool): mount workspaces if True, copy otherwise
     #
-    def _stage_sources_at(self, directory):
+    def _stage_sources_at(self, directory, mount_workspaces=True):
         with self.timed_activity("Staging sources", silent_nested=True):
 
             if os.path.isdir(directory) and os.listdir(directory):
                 raise ElementError("Staging directory '{}' is not empty".format(directory))
 
-            for source in self.__sources:
+            # If mount_workspaces is set, sources with workspace are mounted
+            # directly inside the sandbox so no need to stage them here.
+            if mount_workspaces:
+                sources = [source for source in self.sources() if not source._has_workspace()]
+            else:
+                sources = self.sources()
+
+            for source in sources:
                 source._stage(directory)
 
         # Ensure deterministic mtime of sources at build time
@@ -1458,6 +1474,127 @@ class Element(Plugin):
         context = self._get_context()
         return context._get_strict(project.name)
 
+    # _pull_pending()
+    #
+    # Check whether the artifact will be pulled.
+    #
+    # Returns:
+    #   (bool): Whether a pull operation is pending
+    #
+    def _pull_pending(self):
+        if self.__pull_failed:
+            # Consider this equivalent to artifact being unavailable in
+            # remote cache
+            return False
+
+        if not self.__strong_cached and self.__remotely_strong_cached:
+            # Pull pending using strict cache key
+            return True
+        elif not self.__cached and self.__remotely_cached:
+            # Pull pending using weak cache key
+            return True
+        else:
+            # No pull pending
+            return False
+
+    # _pull_failed()
+    #
+    # Indicate that pull was attempted but failed.
+    #
+    def _pull_failed(self):
+        self.__pull_failed = True
+
+    # _update_state()
+    #
+    # Keep track of element state. Calculate cache keys if possible and
+    # check whether artifacts are cached.
+    #
+    # This must be called whenever the state of an element may have changed.
+    #
+    def _update_state(self):
+        # Determine consistency of sources
+        for source in self.__sources:
+            source._update_state()
+
+        if self._consistency() == Consistency.INCONSISTENT:
+            # Tracking is still pending
+            return
+
+        if any([not source._stable() for source in self.__sources]):
+            # If any source is not stable, discard current cache key values
+            # as their correct values can only be calculated once the build is complete
+            self.__cache_key_dict = None
+            self.__cache_key = None
+            self.__weak_cache_key = None
+            self.__strict_cache_key = None
+            self.__strong_cached = None
+            self.__remotely_cached = None
+            self.__remotely_strong_cached = None
+            return
+
+        if self.__weak_cache_key is None:
+            # Calculate weak cache key
+            # Weak cache key includes names of direct build dependencies
+            # but does not include keys of dependencies.
+            if self.BST_STRICT_REBUILD:
+                dependencies = [
+                    e._get_cache_key(strength=_KeyStrength.WEAK)
+                    for e in self.dependencies(Scope.BUILD)
+                ]
+            else:
+                dependencies = [
+                    e.name for e in self.dependencies(Scope.BUILD, recurse=False)
+                ]
+
+            self.__weak_cache_key = self.__calculate_cache_key(dependencies)
+
+            if self.__weak_cache_key is None:
+                # Weak cache key could not be calculated yet
+                return
+
+        if self.__strict_cache_key is None:
+            dependencies = [
+                e.__strict_cache_key for e in self.dependencies(Scope.BUILD)
+            ]
+            self.__strict_cache_key = self.__calculate_cache_key(dependencies)
+
+            if self.__strict_cache_key is None:
+                # Strict cache key could not be calculated yet
+                return
+
+        # Query caches now that the weak and strict cache keys are available
+        if not self.__cached:
+            self.__cached = self.__artifacts.contains(self)
+        if not self.__remotely_cached:
+            self.__remotely_cached = self.__artifacts.remote_contains(self)
+        if not self.__strong_cached:
+            self.__strong_cached = self.__artifacts.contains(self, strength=_KeyStrength.STRONG)
+        if not self.__remotely_strong_cached:
+            self.__remotely_strong_cached = self.__artifacts.remote_contains(self, strength=_KeyStrength.STRONG)
+
+        if self.__cache_key is None:
+            # Calculate strong cache key
+            if self._get_strict():
+                self.__cache_key = self.__strict_cache_key
+            elif self._pull_pending():
+                # Effective strong cache key is unknown until after the pull
+                pass
+            elif self._cached():
+                # Load the strong cache key from the artifact
+                metadir = os.path.join(self.__artifacts.extract(self), 'meta')
+                meta = _yaml.load(os.path.join(metadir, 'artifact.yaml'))
+                self.__cache_key = meta['keys']['strong']
+            elif self._buildable():
+                # Artifact will be built, not downloaded
+                dependencies = [
+                    e._get_cache_key() for e in self.dependencies(Scope.BUILD)
+                ]
+                self.__cache_key = self.__calculate_cache_key(dependencies)
+
+            if self.__cache_key is None:
+                # Strong cache key could not be calculated yet
+                return
+
     #############################################################
     #                   Private Local Methods                   #
     #############################################################
@@ -1621,6 +1758,20 @@ class Element(Plugin):
 
         return element_public
 
+    def __file_is_whitelisted(self, pattern):
+        # Considered storing the whitelist regex for re-use, but public data
+        # can be altered mid-build.
+        # Public data is not guaranteed to stay the same for the duration of
+        # the build, but I can think of no reason to change it mid-build.
+        # If this ever changes, things will go wrong unexpectedly.
+        if not self.__whitelist_regex:
+            bstdata = self.get_public_data('bst')
+            whitelist = _yaml.node_get(bstdata, list, 'overlap-whitelist', default_value=[])
+            whitelist_expressions = [utils._glob2re(self.__variables.subst(exp.strip())) for exp in whitelist]
+            expression = ('^(?:' + '|'.join(whitelist_expressions) + ')$')
+            self.__whitelist_regex = re.compile(expression)
+        return self.__whitelist_regex.match(pattern)
+
     def __init_splits(self):
         bstdata = self.get_public_data('bst')
         splits = bstdata.get('split-rules')
@@ -1681,9 +1832,22 @@ class Element(Plugin):
                 yield filename.lstrip(os.sep)
 
     def _load_public_data(self):
-        self._assert_cached(recalculate=False)
+        self._assert_cached()
         assert(self.__dynamic_public is None)
 
         # Load the public data from the artifact
         metadir = os.path.join(self.__artifacts.extract(self), 'meta')
         self.__dynamic_public = _yaml.load(os.path.join(metadir, 'public.yaml'))
+
+    def _subst_string(self, value):
+        return self.__variables.subst(value)
+
+
+def _overlap_error_detail(f, forbidden_overlap_elements, elements):
+    if forbidden_overlap_elements:
+        return ("/{}: {} {} not permitted to overlap other elements, order {} \n"
+                .format(f, " and ".join(forbidden_overlap_elements),
+                        "is" if len(forbidden_overlap_elements) == 1 else "are",
+                        " above ".join(reversed(elements))))
+    else:
+        return ""
