@@ -58,11 +58,12 @@ class SourceError(BstError):
     to report errors to the user.
 
     Args:
-       message (str): The error message to report to the user
+       message (str): The breif error description to report to the user
+       detail (str): A possibly multiline, more detailed error message
        reason (str): An optional machine readable reason string, used for test cases
     """
-    def __init__(self, message, *, reason=None):
-        super().__init__(message, domain=ErrorDomain.SOURCE, reason=reason)
+    def __init__(self, message, *, detail=None, reason=None):
+        super().__init__(message, detail=detail, domain=ErrorDomain.SOURCE, reason=reason)
 
 
 class Source(Plugin):
@@ -81,7 +82,9 @@ class Source(Plugin):
         self.__origin_node = meta.origin_node           # YAML node this Source was loaded from
         self.__origin_toplevel = meta.origin_toplevel   # Toplevel YAML node for the file
         self.__origin_filename = meta.origin_filename   # Filename of the file the source was loaded from
-        self.__consistency = None                       # Cached consistency state
+        self.__consistency = Consistency.INCONSISTENT   # Cached consistency state
+        self.__tracking = False                         # Source is scheduled to be tracked
+        self.__assemble_scheduled = False               # Source is scheduled to be assembled
         self.__workspace = None                         # Directory of the currently active workspace
         self.__workspace_key = None                     # Cached directory content hashes for workspaced source
 
@@ -257,10 +260,15 @@ class Source(Plugin):
     #            Private Methods used in BuildStream            #
     #############################################################
 
-    # Wrapper for get_consistency() api which caches the result
+    # Update cached consistency for a source
     #
-    def _get_consistency(self, recalculate=False):
-        if recalculate or self.__consistency is None:
+    # This must be called whenever the state of a source may have changed.
+    #
+    def _update_state(self):
+        if self.__tracking:
+            return
+
+        if self.__consistency < Consistency.CACHED:
             self.__consistency = self.get_consistency()
 
             if self._has_workspace() and \
@@ -273,6 +281,9 @@ class Source(Plugin):
                 if not os.path.exists(fullpath):
                     self.__consistency = Consistency.INCONSISTENT
 
+    # Return cached consistency
+    #
+    def _get_consistency(self):
         return self.__consistency
 
     # Return the absolute path of the element's workspace
@@ -280,16 +291,7 @@ class Source(Plugin):
     def _get_workspace_path(self):
         return os.path.join(self.get_project_directory(), self.__workspace)
 
-    # Bump local cached consistency state, this is done from
-    # the pipeline after the successful completion of fetch
-    # and track jobs.
-    #
-    def _bump_consistency(self, consistency):
-        if (self.__consistency is None or
-            consistency > self.__consistency):
-            self.__consistency = consistency
-
-    # Force a source to appear to be in an inconsistent state.
+    # Mark a source as scheduled to be tracked
     #
     # This is used across the pipeline in sessions where the
     # source in question are going to be tracked. This is important
@@ -297,19 +299,53 @@ class Source(Plugin):
     # keys until the source is RESOLVED and also prevent depending
     # elements from being assembled until the source is CACHED.
     #
-    def _force_inconsistent(self):
-        self.__consistency = Consistency.INCONSISTENT
+    def _schedule_tracking(self):
+        self.__tracking = True
+
+    # _schedule_assemble():
+    #
+    # This is called in the main process before the element is assembled
+    # in a subprocess.
+    #
+    def _schedule_assemble(self):
+        assert(not self.__assemble_scheduled)
+        self.__assemble_scheduled = True
+
+        # Invalidate workspace key as the build modifies the workspace directory
+        self.__workspace_key = None
+
+    # _assemble_done():
+    #
+    # This is called in the main process after the element has been assembled
+    # in a subprocess.
+    #
+    def _assemble_done(self):
+        assert(self.__assemble_scheduled)
+        self.__assemble_scheduled = False
+
+    # _stable():
+    #
+    # Unstable sources are mounted read/write and thus cannot produce a
+    # (stable) cache key before the build is complete.
+    #
+    def _stable(self):
+        # Source directory is modified by workspace build process
+        return not (self._has_workspace() and self.__assemble_scheduled)
 
     # Wrapper function around plugin provided fetch method
     #
     def _fetch(self):
         self.fetch()
 
-    # Ensures a fully constructed path and returns it
-    def _ensure_directory(self, directory):
+    # Return the path where this source should be staged under given directory
+    def _get_staging_path(self, directory):
         if self.__directory is not None:
             directory = os.path.join(directory, self.__directory.lstrip(os.sep))
+        return directory
 
+    # Ensures a fully constructed path and returns it
+    def _ensure_directory(self, directory):
+        directory = self._get_staging_path(directory)
         try:
             os.makedirs(directory, exist_ok=True)
         except OSError as e:
@@ -323,12 +359,12 @@ class Source(Plugin):
     # 'directory' option
     #
     def _stage(self, directory):
-        directory = self._ensure_directory(directory)
+        staging_directory = self._ensure_directory(directory)
 
         if self._has_workspace():
-            self._stage_workspace(directory)
+            self._stage_workspace(staging_directory)
         else:
-            self.stage(directory)
+            self.stage(staging_directory)
 
     # Wrapper for init_workspace()
     def _init_workspace(self, directory):
@@ -365,6 +401,8 @@ class Source(Plugin):
             self.set_ref(ref, node)
             changed = True
 
+        self.__tracking = False
+
         return changed
 
     # Wrapper for track()
@@ -373,10 +411,16 @@ class Source(Plugin):
         new_ref = self.track()
         current_ref = self.get_ref()
 
-        # It's consistent unless it reported an error
-        self._bump_consistency(Consistency.RESOLVED)
+        if new_ref is None:
+            # No tracking, keep current ref
+            new_ref = current_ref
+
         if current_ref != new_ref:
             self.info("Found new revision: {}".format(new_ref))
+            if self._has_workspace():
+                detail = "This source has an open workspace.\n" \
+                    + "To start using the new reference, please close the existing workspace."
+                self.warn("Updated reference will be ignored as source has open workspace", detail=detail)
 
         return new_ref
 
@@ -424,6 +468,8 @@ class Source(Plugin):
     # new calculation to happen by setting the 'recalculate' flag.
     #
     def _get_workspace_key(self, recalculate=False):
+        assert(not self.__assemble_scheduled)
+
         if recalculate or self.__workspace_key is None:
             fullpath = self._get_workspace_path()
 
