@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 #
-#  Copyright (C) 2016 Codethink Limited
+#  Copyright (C) 2016-2018 Codethink Limited
 #
 #  This program is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU Lesser General Public
@@ -26,7 +26,6 @@ import calendar
 import errno
 import hashlib
 import os
-import pickle
 import re
 import shutil
 import signal
@@ -34,13 +33,14 @@ import stat
 import string
 import subprocess
 import tempfile
+import itertools
+import functools
 from contextlib import contextmanager
 
-import pkg_resources
 import psutil
 
+from . import __version__
 from . import _signals
-from . import _yaml
 from ._exceptions import BstError, ErrorDomain
 
 
@@ -211,6 +211,7 @@ def sha256sum(filename):
         with open(filename, "rb") as f:
             for chunk in iter(lambda: f.read(4096), b""):
                 h.update(chunk)
+
     except OSError as e:
         raise UtilError("Failed to get a checksum of file '{}': {}"
                         .format(filename, e)) from e
@@ -455,12 +456,69 @@ def get_bst_version():
        (int): The major version
        (int): The minor version
     """
-    package = pkg_resources.require("buildstream")[0]
-    versions = package.version.split('.')[:2]
+    versions = __version__.split('.')[:2]
 
     return (int(versions[0]), int(versions[1]))
 
 
+@contextmanager
+def save_file_atomic(filename, mode='w', *, buffering=-1, encoding=None,
+                     errors=None, newline=None, closefd=True, opener=None):
+    """Save a file with a temporary name and rename it into place when ready.
+
+    This is a context manager which is meant for saving data to files.
+    The data is written to a temporary file, which gets renamed to the target
+    name when the context is closed. This avoids readers of the file from
+    getting an incomplete file.
+
+    **Example:**
+
+    .. code:: python
+
+      with save_file_atomic('/path/to/foo', 'w') as f:
+          f.write(stuff)
+
+    The file will be called something like ``tmpCAFEBEEF`` until the
+    context block ends, at which point it gets renamed to ``foo``. The
+    temporary file will be created in the same directory as the output file.
+    The ``filename`` parameter must be an absolute path.
+
+    If an exception occurs or the process is terminated, the temporary file will
+    be deleted.
+    """
+    # This feature has been proposed for upstream Python in the past, e.g.:
+    # https://bugs.python.org/issue8604
+
+    assert os.path.isabs(filename), "The utils.save_file_atomic() parameter ``filename`` must be an absolute path"
+    dirname = os.path.dirname(filename)
+    fd, tempname = tempfile.mkstemp(dir=dirname)
+    os.close(fd)
+
+    f = open(tempname, mode=mode, buffering=buffering, encoding=encoding,
+             errors=errors, newline=newline, closefd=closefd, opener=opener)
+
+    def cleanup_tempfile():
+        f.close()
+        try:
+            os.remove(tempname)
+        except FileNotFoundError:
+            pass
+        except OSError as e:
+            raise UtilError("Failed to cleanup temporary file {}: {}".format(tempname, e)) from e
+
+    try:
+        with _signals.terminator(cleanup_tempfile):
+            f.real_filename = filename
+            yield f
+            f.close()
+            # This operation is atomic, at least on platforms we care about:
+            # https://bugs.python.org/issue8828
+            os.replace(tempname, filename)
+    except Exception as e:
+        cleanup_tempfile()
+        raise
+
+
 # Recursively remove directories, ignoring file permissions as much as
 # possible.
 def _force_rmtree(rootpath, **kwargs):
@@ -503,6 +561,11 @@ def _copy_directories(srcdir, destdir, target):
                                 'directory expected: {}'.format(old_dir))
 
 
+@functools.lru_cache(maxsize=64)
+def _resolve_symlinks(path):
+    return os.path.realpath(path)
+
+
 def _ensure_real_directory(root, destpath):
     # The realpath in the sandbox may refer to a file outside of the
     # sandbox when any of the direcory branches are a symlink to an
@@ -511,12 +574,12 @@ def _ensure_real_directory(root, destpath):
     # This should not happen as we rely on relative_symlink_target() below
     # when staging the actual symlinks which may lead up to this path.
     #
-    realpath = os.path.realpath(destpath)
-    if not realpath.startswith(os.path.realpath(root)):
+    destpath_resolved = _resolve_symlinks(destpath)
+    if not destpath_resolved.startswith(_resolve_symlinks(root)):
         raise UtilError('Destination path resolves to a path outside ' +
                         'of the staging area\n\n' +
                         '  Destination path: {}\n'.format(destpath) +
-                        '  Real path: {}'.format(realpath))
+                        '  Real path: {}'.format(destpath_resolved))
 
     # Ensure the real destination path exists before trying to get the mode
     # of the real destination path.
@@ -525,10 +588,10 @@ def _ensure_real_directory(root, destpath):
     # refer to non-existing directories, they will be created on demand here
     # at staging time.
     #
-    if not os.path.exists(realpath):
-        os.makedirs(realpath)
+    if not os.path.exists(destpath_resolved):
+        os.makedirs(destpath_resolved)
 
-    return realpath
+    return destpath_resolved
 
 
 # _process_list()
@@ -599,7 +662,7 @@ def _process_list(srcdir, destdir, filelist, actionfunc, result,
             if not os.path.exists(destpath):
                 _ensure_real_directory(destdir, destpath)
 
-            dest_stat = os.lstat(os.path.realpath(destpath))
+            dest_stat = os.lstat(_resolve_symlinks(destpath))
             if not stat.S_ISDIR(dest_stat.st_mode):
                 raise UtilError('Destination not a directory. source has {}'
                                 ' destination has {}'.format(srcpath, destpath))
@@ -671,11 +734,11 @@ def _relative_symlink_target(root, symlink, target):
 
         # We want a relative path from the directory in which symlink
         # is located, not from the symlink itself.
-        symlinkdir, _ = os.path.split(os.path.realpath(symlink))
+        symlinkdir, _ = os.path.split(_resolve_symlinks(symlink))
 
         # Create a full path to the target, including the leading staging
         # directory
-        fulltarget = os.path.join(os.path.realpath(root), target)
+        fulltarget = os.path.join(_resolve_symlinks(root), target)
 
         # now get the relative path from the directory where the symlink
         # is located within the staging root, to the target within the same
@@ -687,24 +750,6 @@ def _relative_symlink_target(root, symlink, target):
         return target
 
 
-# _generate_key()
-#
-# Generate an sha256 hex digest from the given value. The value
-# can be a simple value or recursive dictionary with lists etc,
-# anything simple enough to serialize.
-#
-# Args:
-#    value: A value to get a key for
-#
-# Returns:
-#    (str): An sha256 hex digest of the given value
-#
-def _generate_key(value):
-    ordered = _yaml.node_sanitize(value)
-    string = pickle.dumps(ordered)
-    return hashlib.sha256(string).hexdigest()
-
-
 # _set_deterministic_user()
 #
 # Set the uid/gid for every file in a directory tree to the process'
@@ -954,3 +999,32 @@ def _glob2re(pat):
         else:
             res = res + re.escape(c)
     return res + '\Z(?ms)'
+
+
+# _deduplicate()
+#
+# Remove duplicate entries in a list or other iterable.
+#
+# Copied verbatim from the unique_everseen() example at
+# https://docs.python.org/3/library/itertools.html#itertools-recipes
+#
+# Args:
+#    iterable (iterable): What to deduplicate
+#    key (callable): Optional function to map from list entry to value
+#
+# Returns:
+#    (generator): Generator that produces a deduplicated version of 'iterable'
+#
+def _deduplicate(iterable, key=None):
+    seen = set()
+    seen_add = seen.add
+    if key is None:
+        for element in itertools.filterfalse(seen.__contains__, iterable):
+            seen_add(element)
+            yield element
+    else:
+        for element in iterable:
+            k = key(element)
+            if k not in seen:
+                seen_add(k)
+                yield element
