@@ -1,9 +1,11 @@
 import os
 import pytest
 import shutil
+import subprocess
 from tests.testutils import cli, create_repo, ALL_REPO_KINDS
 
 from buildstream import _yaml
+from buildstream._exceptions import ErrorDomain, LoadErrorReason
 
 repo_kinds = [(kind) for kind in ALL_REPO_KINDS]
 
@@ -74,6 +76,24 @@ def test_open(cli, tmpdir, datafiles, kind):
     open_workspace(cli, tmpdir, datafiles, kind, False)
 
 
+@pytest.mark.datafiles(DATA_DIR)
+def test_open_bzr(cli, tmpdir, datafiles):
+    element_name, project, workspace = open_workspace(cli, tmpdir, datafiles, "bzr", False)
+
+    # Check that the .bzr dir exists
+    bzrdir = os.path.join(workspace, ".bzr")
+    assert(os.path.isdir(bzrdir))
+
+    # Check that the correct origin branch is set
+    element_config = _yaml.load(os.path.join(project, "elements", element_name))
+    source_config = element_config['sources'][0]
+    output = subprocess.check_output(["bzr", "info"], cwd=workspace)
+    stripped_url = source_config['url'].lstrip("file:///")
+    expected_output_str = ("checkout of branch: /{}/{}"
+                           .format(stripped_url, source_config['track']))
+    assert(expected_output_str in str(output))
+
+
 @pytest.mark.datafiles(DATA_DIR)
 @pytest.mark.parametrize("kind", repo_kinds)
 def test_open_track(cli, tmpdir, datafiles, kind):
@@ -189,3 +209,56 @@ def test_build(cli, tmpdir, datafiles, kind):
 
     # Check that the original /usr/bin/hello is not in the checkout
     assert not os.path.exists(os.path.join(workspace, 'usr', 'bin', 'hello'))
+
+
+@pytest.mark.datafiles(DATA_DIR)
+def test_open_old_format(cli, tmpdir, datafiles):
+    project = os.path.join(datafiles.dirname, datafiles.basename)
+    os.makedirs(os.path.join(project, '.bst'))
+    workspace_config_path = os.path.join(project, '.bst', 'workspaces.yml')
+
+    workspace_dict_old = {
+        "alpha.bst": {
+            0: "/workspaces/bravo",
+        }
+    }
+
+    workspace_dict_expected = {
+        "alpha.bst": "/workspaces/bravo"
+    }
+
+    _yaml.dump(workspace_dict_old, workspace_config_path)
+
+    # Check that we can still read workspace config that is in old format
+    result = cli.run(project=project, args=['workspace', 'list'])
+    result.assert_success()
+
+    loaded_config = _yaml.load(workspace_config_path)
+
+    # Check that workspace config has been converted to the new format after
+    # running a workspace command
+    elements = list(_yaml.node_items(loaded_config))
+    assert len(elements) == len(workspace_dict_expected)
+    for element, path in workspace_dict_expected.items():
+        assert path == _yaml.node_get(loaded_config, str, element)
+
+
+@pytest.mark.datafiles(DATA_DIR)
+def test_open_old_format_fail(cli, tmpdir, datafiles):
+    project = os.path.join(datafiles.dirname, datafiles.basename)
+    os.makedirs(os.path.join(project, '.bst'))
+    workspace_config_path = os.path.join(project, '.bst', 'workspaces.yml')
+
+    workspace_dict_old = {
+        "alpha.bst": {
+            0: "/workspaces/bravo",
+            1: "/workspaces/charlie",
+        }
+    }
+
+    _yaml.dump(workspace_dict_old, workspace_config_path)
+
+    # Check that trying to load workspace config in old format raises an error
+    # if there are workspaces open for more than one source for an element
+    result = cli.run(project=project, args=['workspace', 'list'])
+    result.assert_main_error(ErrorDomain.LOAD, LoadErrorReason.INVALID_DATA)
